@@ -188,9 +188,9 @@ handle_exepipe (GIOChannel * source,
                         retval = TRUE;
                 }
                 g_string_free (str, TRUE);
-        } else {                
-                retval = FALSE;
         }
+        else
+                retval = FALSE;
 
 	if (retval && !strncmp (data->iorbuf, "IOR:", 4))
 		retval = FALSE;
@@ -208,7 +208,7 @@ handle_exepipe (GIOChannel * source,
 
 CORBA_Object
 bonobo_activation_server_by_forking (
-	const char                         **cmd,
+	const char                         **cmd_const,
 	gboolean                             set_process_group,
 	int                                  fd_arg, 
 	const Bonobo_ActivationEnvironment  *environment,
@@ -228,6 +228,7 @@ bonobo_activation_server_by_forking (
         GSource *source;
         GMainContext *context;
         char **newenv = NULL;
+        char **cmd;
 
 #if defined(__APPLE__) && defined(HAVE_NSGETENVIRON) && defined(HAVE_CRT_EXTERNS_H)
 # include <crt_externs.h>
@@ -238,13 +239,13 @@ bonobo_activation_server_by_forking (
         extern char **environ;
 #endif
 
-        g_return_val_if_fail (cmd != NULL, CORBA_OBJECT_NIL);
-        g_return_val_if_fail (cmd [0] != NULL, CORBA_OBJECT_NIL);
+        g_return_val_if_fail (cmd_const != NULL, CORBA_OBJECT_NIL);
+        g_return_val_if_fail (cmd_const [0] != NULL, CORBA_OBJECT_NIL);
         g_return_val_if_fail (act_iid != NULL, CORBA_OBJECT_NIL);
 
         ai.environment = environment;
         ai.act_iid = act_iid;
-        ai.exename = cmd [0];
+        ai.exename = cmd_const [0];
         ai.re_check = re_check;
         ai.user_data = user_data;
 
@@ -260,7 +261,7 @@ bonobo_activation_server_by_forking (
 
 #ifdef BONOBO_ACTIVATION_DEBUG
         if (ai.do_srv_output)
-                fprintf (stderr, " SPAWNING: '%s' for '%s'\n", cmd[0], act_iid);
+                fprintf (stderr, " SPAWNING: '%s' for '%s'\n", cmd_const[0], act_iid);
 #endif
 
 #ifdef G_OS_WIN32
@@ -322,9 +323,13 @@ bonobo_activation_server_by_forking (
                 newenv [n] = NULL;
         }
 
+        cmd = g_strdupv (cmd_const);
         /* Pass the IOR pipe's write end to the child */ 
         if (fd_arg != 0)
-                cmd[fd_arg] = g_strdup_printf (cmd[fd_arg], iopipes[1]);
+        {
+                g_free (cmd[fd_arg]);
+                cmd[fd_arg] = g_strdup_printf (cmd_const[fd_arg], iopipes[1]);
+        }
 
         ai.iorbuf[0] = '\0';
         ai.done = FALSE;
@@ -364,10 +369,7 @@ bonobo_activation_server_by_forking (
                 if (use_new_loop)
                         g_main_context_unref (context);
 
-                if (fd_arg != 0) {
-                        g_free ((char *) cmd[fd_arg]);
-                        cmd[fd_arg] = NULL;
-                }
+                g_strfreev (cmd);
 
 		return CORBA_OBJECT_NIL;
 	}
@@ -383,11 +385,7 @@ bonobo_activation_server_by_forking (
                 }
                 g_free (newenv);
         }
-
-        if (fd_arg != 0) {
-                g_free ((char *) cmd[fd_arg]);
-                cmd[fd_arg] = NULL;
-        }
+        g_strfreev (cmd);
 
         /* Get the IOR from the pipe */
         while (!ai.done) {
