@@ -17,7 +17,7 @@
 #include <bonobo/bonobo-event-source.h>
 #include <bonobo/bonobo-moniker-util.h>
 #include <bonobo/bonobo-running-context.h>
-#include <bonobo/bonobo-shutdown.h>
+#include <bonobo/bonobo-private.h>
 #include <bonobo/bonobo-main.h>
 #include <bonobo/bonobo-debug.h>
 
@@ -51,6 +51,7 @@ key_free (gpointer name, gpointer dummy1, gpointer user_data)
 	g_free (name);
 }
 
+#ifdef G_ENABLE_DEBUG
 static void
 bonobo_ri_debug_foreach (gpointer key, gpointer value, gpointer user_data)
 {
@@ -59,6 +60,7 @@ bonobo_ri_debug_foreach (gpointer key, gpointer value, gpointer user_data)
 	bonobo_debug_print ("", "[%p]:CORBA_Object still running", o);
 		
 }
+#endif
 
 void
 bonobo_running_context_shutdown (void)
@@ -105,7 +107,7 @@ check_destroy (GObject *object)
 }
 
 static BonoboRunningInfo *
-get_running_info (gboolean create)
+get_running_info_T (gboolean create)
 {
 	if (!bonobo_running_info && create) {
 		bonobo_running_info = g_new (BonoboRunningInfo, 1);
@@ -118,9 +120,9 @@ get_running_info (gboolean create)
 }
 
 static void
-check_empty (void)
+check_empty_T (void)
 {
-	BonoboRunningInfo *ri = get_running_info (FALSE);
+	BonoboRunningInfo *ri = get_running_info_T (FALSE);
 
 	if (!ri || !bonobo_running_context)
 		return;
@@ -130,52 +132,55 @@ check_empty (void)
 	    (g_hash_table_size (ri->keys) == 0)) {
 
 		ri->emitted_last_unref = TRUE;
+		BONOBO_UNLOCK ();
 
 		g_signal_emit (G_OBJECT (bonobo_running_context),
 			       signals [LAST_UNREF], 0);
-
 		g_return_if_fail (bonobo_running_event_source != NULL);
 
 		bonobo_event_source_notify_listeners (
 			bonobo_running_event_source,
 			"bonobo:last_unref", NULL, NULL);
+
+		BONOBO_LOCK ();
 	}
 }
 
-#ifndef bonobo_running_context_add_object
+#ifndef bonobo_running_context_add_object_T
 void
-bonobo_running_context_add_object (CORBA_Object object)
+bonobo_running_context_add_object_T (CORBA_Object object)
 {
 #ifdef G_ENABLE_DEBUG
 	if(_bonobo_debug_flags & BONOBO_DEBUG_RUNNING)
-		bonobo_running_context_trace_objects (object, "local", 0, 0);
+		bonobo_running_context_trace_objects_T (object, "local", 0, 0);
 	else
 #endif /* G_ENABLE_DEBUG */
 	{
-		BonoboRunningInfo *ri = get_running_info (TRUE);
+		BonoboRunningInfo *ri = get_running_info_T (TRUE);
 
+		ri->emitted_last_unref = FALSE;
 		g_hash_table_insert (ri->objects, object, object);
 	}
 }
 #endif
 
 
-#ifndef bonobo_running_context_remove_object
+#ifndef bonobo_running_context_remove_object_T
 void
-bonobo_running_context_remove_object (CORBA_Object object)
+bonobo_running_context_remove_object_T (CORBA_Object object)
 {
 #ifdef G_ENABLE_DEBUG
 	if(_bonobo_debug_flags & BONOBO_DEBUG_RUNNING)
-		bonobo_running_context_trace_objects (object, "local", 0, 1);
+		bonobo_running_context_trace_objects_T (object, "local", 0, 1);
 	else
 #endif /* G_ENABLE_DEBUG */
 	{
-		BonoboRunningInfo *ri = get_running_info (FALSE);
+		BonoboRunningInfo *ri = get_running_info_T (FALSE);
 
 		if (ri) {
 			g_hash_table_remove (ri->objects, object);
 
-			check_empty ();
+			check_empty_T ();
 		}
 	}
 }
@@ -185,43 +190,68 @@ bonobo_running_context_remove_object (CORBA_Object object)
 void
 bonobo_running_context_ignore_object (CORBA_Object object)
 {
+	BONOBO_LOCK ();
 #ifdef G_ENABLE_DEBUG
 	if(_bonobo_debug_flags & BONOBO_DEBUG_RUNNING)
-		bonobo_running_context_trace_objects (object, "local", 0, 2);
+		bonobo_running_context_trace_objects_T (object, "local", 0, 2);
 	else
 #endif /* G_ENABLE_DEBUG */
 	{
-		BonoboRunningInfo *ri = get_running_info (FALSE);
+		BonoboRunningInfo *ri = get_running_info_T (FALSE);
 
-		if (ri) {
+		if (ri)
 			g_hash_table_remove (ri->objects, object);
-		}
 	}
+	BONOBO_UNLOCK ();
 }
 #endif
 
-void          
-bonobo_running_context_trace_objects (CORBA_Object object,
-				      const char  *fn,
-				      int          line,
-				      int          mode)
+static void
+_running_context_list_objects (gpointer key,
+			       gpointer value,
+			       gpointer user_data)
 {
-	BonoboRunningInfo *ri = get_running_info (mode == 0);
+	CORBA_Object object = (CORBA_Object) value;
+	CORBA_char *type_id;
+	CORBA_Environment ev;
+
+	CORBA_exception_init (&ev);
+	type_id = ORBit_small_get_type_id (object, &ev);
+	if (ev._major != CORBA_NO_EXCEPTION)
+		type_id = CORBA_string_dup (
+			"<error determining object type id!>");
+	CORBA_exception_free (&ev);
+
+	bonobo_debug_print ("Alive: ", "[%p]: %s", object, type_id);
+	CORBA_free (type_id);
+}
+
+void          
+bonobo_running_context_trace_objects_T (CORBA_Object object,
+					const char  *fn,
+					int          line,
+					int          mode)
+{
+	BonoboRunningInfo *ri;
+#ifdef G_ENABLE_DEBUG
 	static char *cmode[] = {
 		"add_object",
 		"remove_object",
 		"ignore_object"		
 	};
+#endif
+	ri = get_running_info_T (mode == 0);	
 
 	if (ri) {
 		switch (mode) {
 		case 0:
 			g_hash_table_insert (ri->objects, object, object);
+			ri->emitted_last_unref = FALSE;
 			break;
 		case 1:
 			g_hash_table_remove (ri->objects, object);
 
-			check_empty ();
+			check_empty_T ();
 			break;
 		case 2:
 			g_hash_table_remove (ri->objects, object);
@@ -229,10 +259,12 @@ bonobo_running_context_trace_objects (CORBA_Object object,
 		}
 
 #ifdef G_ENABLE_DEBUG
-		if(_bonobo_debug_flags & BONOBO_DEBUG_RUNNING)
+		if(_bonobo_debug_flags & BONOBO_DEBUG_RUNNING) {
 			bonobo_debug_print (cmode [mode], 
 					    "[%p]:CORBA_Object %d running objects at %s:%d",
 					    object, g_hash_table_size (ri->objects), fn, line);
+			g_hash_table_foreach (ri->objects, _running_context_list_objects, NULL);
+		}
 #endif /* G_ENABLE_DEBUG */
 	}
 }
@@ -242,7 +274,9 @@ impl_Bonobo_RunningContext_addObject (PortableServer_Servant servant,
 				      const CORBA_Object     object,
 				      CORBA_Environment     *ev)
 {
-	bonobo_running_context_add_object (object);
+	BONOBO_LOCK ();
+	bonobo_running_context_add_object_T (object);
+	BONOBO_UNLOCK ();
 }
 
 static void
@@ -250,7 +284,9 @@ impl_Bonobo_RunningContext_removeObject (PortableServer_Servant servant,
 					 const CORBA_Object     object,
 					 CORBA_Environment     *ev)
 {
-	bonobo_running_context_remove_object (object);
+	BONOBO_LOCK ();
+	bonobo_running_context_remove_object_T (object);
+	BONOBO_UNLOCK ();
 }
 
 static void
@@ -259,7 +295,11 @@ impl_Bonobo_RunningContext_addKey (PortableServer_Servant servant,
 				   CORBA_Environment     *ev)
 {
 	char              *key_copy, *old_key;
-	BonoboRunningInfo *ri = get_running_info (TRUE);
+	BonoboRunningInfo *ri;
+
+	BONOBO_LOCK ();
+
+	ri = get_running_info_T (TRUE);
 
 	old_key = g_hash_table_lookup (ri->keys, key);
 	if (old_key) {
@@ -269,6 +309,8 @@ impl_Bonobo_RunningContext_addKey (PortableServer_Servant servant,
 	key_copy = g_strdup (key);
 
 	g_hash_table_insert (ri->keys, key_copy, key_copy);
+
+	BONOBO_UNLOCK ();
 }
 
 static void
@@ -276,18 +318,23 @@ impl_Bonobo_RunningContext_removeKey (PortableServer_Servant servant,
 				      const CORBA_char      *key,
 				      CORBA_Environment     *ev)
 {
-	BonoboRunningInfo *ri = get_running_info (FALSE);
+	BonoboRunningInfo *ri;
 	char              *old_key;
 
-	if (!ri)
-		return;
+	BONOBO_LOCK ();
 
-	old_key = g_hash_table_lookup (ri->keys, key);
-	if (old_key)
-		g_free (old_key);
-	g_hash_table_remove (ri->keys, key);
+	ri = get_running_info_T (FALSE);
 
-	check_empty ();
+	if (ri) {
+		old_key = g_hash_table_lookup (ri->keys, key);
+		if (old_key)
+			g_free (old_key);
+		g_hash_table_remove (ri->keys, key);
+
+		check_empty_T ();
+	}
+
+	BONOBO_UNLOCK ();
 }
 
 static void
